from abc import ABCMeta, abstractmethod
from collections import defaultdict
from copy import deepcopy
from typing import Union, Type, Any, Tuple

import numpy as np
import torch
import torch.nn as nn
from scipy.signal import find_peaks_cwt

from roerich.algorithms.net import MyNN, MyNNRegressor
from roerich.utils import autoregression_matrix, unified_score
from roerich.metrics.metrics import KL_sym, KL, JSD, PE, PE_sym, Wasserstein
from roerich.algorithms.scaler import SmaScalerCache
from roerich.helper import SMA


class ChangePointDetection(metaclass=ABCMeta):
    def __init__(self, scaler: Any = "default", metric: str = "KL", window_size: int = 1, periods: int = 10,
                 lag_size: int = 0, step: int = 1, n_epochs: int = 100, lr: float = 0.01, lam: float = 0,
                 optimizer: str = "Adam", debug: int = 0):
        """
        
        Parameters
        ----------
        scaler: A scaler object is used to scale an input data. The default one is `SmaScalerCache`
        metric: A loss function during optimize step of NN. Can be one of the following KL_sym, KL, JSD, PE, PE_sym, Wasserstein
        window_size: A size of a window when splitting input data into train and test arrays
        periods: A number of previous data-points used when constructing autoregressive matrix
        lag_size: A distance between train- and test- windows
        step: Each `step`-th data-point is used when creating the input dataset
        n_epochs: A number of epochs during training NN
        lr: A learning rate at each step of optimizer
        lam: A regularization rate
        optimizer: One of Adam, SGD, RMSprop or ASGD optimizers
        debug: default zero
        """
        
        self.scaler = SmaScalerCache(window_size + lag_size) if scaler == "default" else scaler
        
        self.metric = metric
        self.window_size = window_size
        self.periods = periods
        self.lag_size = lag_size
        self.step = step
        self.n_epochs = n_epochs
        self.lr = lr
        self.lam = lam
        self.debug = debug
        
        self._time_shift = lag_size + window_size
        self.avg_window = lag_size + window_size
        self.peak_widths = [0.25 * (lag_size + window_size)]
        
        self.optimizers = defaultdict(lambda: torch.optim.Adam)
        
        self.optimizers["Adam"] = torch.optim.Adam
        self.optimizers["SGD"] = torch.optim.SGD
        self.optimizers["RMSprop"] = torch.optim.RMSprop
        self.optimizers["ASGD"] = torch.optim.ASGD
        self.optimizer = self.optimizers[optimizer]
        
        self.metric_func = {"KL_sym": KL_sym,
                            "KL": KL,
                            "JSD": JSD,
                            "PE": PE,
                            "PE_sym": PE_sym,
                            "W": Wasserstein
                            }
    
    @abstractmethod
    def init_net(self, n_inputs: int) -> None:
        """
        Initialize neural network based on `self.base_net` class
        Parameters
        ----------
        n_inputs: Number of inputs of neural network
        -------

        """
        pass
    
    def predict(self, X: Union[np.ndarray, torch.Tensor]) -> Tuple[Any, Any]:
        """
        Determines a CPD score for every data-point
        Parameters
        ----------
        X: An input data

        Returns `avg_unified_score`: An averaged, unified and shifted CPD score for every data-point in X
                `peaks` Locations of CPD points along all data-points
        -------

        """
        X_auto = autoregression_matrix(X, periods=self.periods, fill_value=0)
        self.init_net(X_auto.shape[1])
        T, reference, test = self.reference_test(X_auto)
        scores = []
        for i in range(len(reference)):
            X_, y_ = self.preprocess(reference[i], test[i])
            score = self.reference_test_predict(X_, y_)
            scores.append(score)
        T_scores = np.array([T[i] for i in range(len(reference))])
        scores = np.array(scores)
        
        # todo optimize memory
        T_uni = np.arange(len(X))
        T_scores = T_scores - self._time_shift
        un_score = unified_score(T_uni, T_scores, scores)
        avg_unified_score = SMA(un_score, self.avg_window)
        peaks = self.find_peaks_cwt(avg_unified_score, widths=self.peak_widths)
        
        return avg_unified_score, peaks
    
    def reference_test(self, X: Union[torch.Tensor, np.ndarray]) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Creates reference and test datasets based on autoregressive matrix.
        
        Parameters
        ----------
        X: An autoregressive matrix

        Returns tuple of numpy arrays: time-steps, reference and test datasets
        -------

        """
        N = self.lag_size
        ws = self.window_size
        T = []
        reference = []
        test = []
        for i in range(2 * ws + N - 1, len(X), self.step):
            T.append(i)
            reference.append(X[i - 2 * ws - N + 1:i - ws - N + 1])
            test.append(X[i - ws + 1:i + 1])
        return np.array(T), np.array(reference), np.array(test)
    
    def preprocess(self, X_ref: np.ndarray, X_test: np.ndarray) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Creates X and y datasets for training NN by stacking reference and test datasets.
        Also applies a scaling transformation into resulting X dataset.
        Labels for reference data-points is 1s.
        Labels for test data-points is 0s.
        
        Parameters
        ----------
        X_ref: reference data-points
        X_test: test data-points

        Returns
        -------
        Tuple of training data
        """
        y_ref = np.zeros(len(X_ref))
        y_test = np.ones(len(X_test))
        X = np.vstack((X_ref, X_test))
        y = np.hstack((y_ref, y_test))
        
        X = self.scaler.fit_transform(X)
        
        X = torch.from_numpy(X).float()
        y = torch.from_numpy(y).float()
        return X, y
    
    def find_peaks_cwt(self, vector, *args, **kwargs):
        """
        Find peaks function based on scipy.signal package
        Parameters
        ----------
        vector: CPD scores array
        args: see docs for https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.find_peaks_cwt.html
        kwargs

        Returns
        -------
        Array with location of peaks
        """
        peaks = find_peaks_cwt(vector, *args, **kwargs)
        return peaks
    
    @abstractmethod
    def reference_test_predict(self, X: torch.Tensor, y: torch.Tensor):
        """
        Training process of forward, backward and optimize steps.
        Parameters
        ----------
        X: train data
        y: train labels

        Returns
        -------
        None
        """
        pass


class OnlineNNClassifier(ChangePointDetection):
    def __init__(self, net: Union[Type[nn.Module], str] = "default", *args, **kwargs):
        """
        Parameters
        ----------
        net: Custom torch.nn.Module neural network or "default" one
        args: see parent class
        kwargs: see parent class
        """
        super().__init__(*args, **kwargs)
        self.criterion = nn.BCELoss()
        
        self.base_net = MyNN if net == "default" else net
        self.net = None
        self.opt = None
    
    def init_net(self, n_inputs):
        self.net = self.base_net(n_inputs)
        self.opt = self.optimizer(
            self.net.parameters(),
            lr=self.lr,
            weight_decay=self.lam
        )
    
    def reference_test_predict(self, X, y):
        self.net.train(False)
        n_last = min(self.window_size, self.step)
        ref_preds = self.net(X[y == 0][-n_last:]).detach().numpy()
        test_preds = self.net(X[y == 1][-n_last:]).detach().numpy()
        
        self.net.train(True)
        for epoch in range(self.n_epochs):  # loop over the dataset multiple times
            # forward + backward + optimize
            outputs = self.net(X)
            loss = self.criterion(outputs.squeeze(), y)
            
            # set gradients to zero
            self.opt.zero_grad()
            loss.backward()
            self.opt.step()
        
        score = self.metric_func[self.metric](ref_preds, test_preds)
        return score


class OnlineNNRuLSIF(ChangePointDetection):
    
    def __init__(self, alpha, net="default", *args, **kwargs):
        """
        Parameters
        ----------
        alpha: The `alpha` parameter in a loss function
        net: Custom torch.nn.Module neural network or "default" one
        args: see parent class
        kwargs: see parent class
        """
        super().__init__(*args, **kwargs)
        
        self.alpha = alpha
        self.base_net = MyNNRegressor if net == "default" else net
        self.net1 = None
        self.net2 = None
        self.opt1 = None
        self.opt2 = None
    
    def init_net(self, n_inputs):
        self.net1 = self.base_net(n_inputs)
        self.opt1 = self.optimizer(
            self.net1.parameters(),
            lr=self.lr,
            weight_decay=self.lam
        )
        self.net2 = deepcopy(self.net1)
        self.opt2 = deepcopy(self.opt1)
    
    def compute_loss(self, y_pred_batch_ref, y_pred_batch_test):
        loss = 0.5 * (1 - self.alpha) * (y_pred_batch_ref ** 2).mean() + \
               0.5 * self.alpha * (y_pred_batch_test ** 2).mean() - (y_pred_batch_test).mean()
        
        return loss
    
    def reference_test_predict(self, X, y):
        n_last = min(self.window_size, self.step)
        self.net1.train(False)
        test_preds = self.net1(X[y == 1][-n_last:]).detach().numpy()
        
        self.net2.train(False)
        ref_preds = self.net2(X[y == 0][-n_last:]).detach().numpy()
        
        self.net1.train(True)
        self.net2.train(True)
        for epoch in range(self.n_epochs):  # loop over the dataset multiple times
            
            # forward + backward + optimize
            y_pred_batch = self.net1(X).squeeze()
            y_pred_batch_ref = y_pred_batch[y == 0]
            y_pred_batch_test = y_pred_batch[y == 1]
            loss1 = self.compute_loss(y_pred_batch_ref, y_pred_batch_test)
            
            # set gradients to zero
            self.opt1.zero_grad()
            loss1.backward()
            self.opt1.step()
            
            # forward + backward + optimize
            y_pred_batch = self.net2(X).squeeze()
            y_pred_batch_ref = y_pred_batch[y == 1]
            y_pred_batch_test = y_pred_batch[y == 0]
            loss2 = self.compute_loss(y_pred_batch_ref, y_pred_batch_test)
            
            # set gradients to zero
            self.opt2.zero_grad()
            loss2.backward()
            self.opt2.step()
        
        score = (0.5 * np.mean(test_preds) - 0.5) + (0.5 * np.mean(ref_preds) - 0.5)
        return score



import numpy as np
from sklearn.metrics import roc_curve, roc_auc_score


def KL_score_unsym(ref_ratios, test_ratios):
    score = np.mean(np.log(test_ratios))
    return score


def KL_score(ref_ratios, test_ratios):
    score = KL_score_unsym(ref_ratios, test_ratios) + KL_score_unsym(1. / test_ratios, 1. / ref_ratios)
    return score


def PE_score_unsym(ref_ratios, test_ratios, alpha=0.):
    score = (-0.5 * alpha * np.mean(test_ratios ** 2)) + \
            (-0.5 * (1. - alpha) * np.mean(ref_ratios ** 2)) + np.mean(test_ratios) - 0.5
    return score


def PE_score(ref_ratios, test_ratios, alpha=0.):
    score = PE_score_unsym(ref_ratios, test_ratios, alpha)  # - PE_score_unsym(test_ratios, ref_ratios, alpha)
    return score


def KL(ref_preds, test_preds):
    return np.mean(np.log(test_preds + 10 ** -3)) - np.mean(np.log(1. - test_preds + 10 ** -3))


def KL_sym(ref_preds, test_preds):
    return np.mean(np.log(test_preds + 10 ** -3)) - np.mean(np.log(1. - test_preds + 10 ** -3)) + \
           np.mean(np.log(1. - ref_preds + 10 ** -3)) - np.mean(np.log(ref_preds + 10 ** -3))


def JSD(ref_preds, test_preds):
    return np.log(2) + 0.5 * np.mean(np.log(test_preds + 10 ** -3)) + 0.5 * np.mean(np.log(1. - ref_preds + 10 ** -3))


def PE(ref_preds, test_preds):
    scores = test_preds / (1. - test_preds + 10 ** -6) - 1.
    scores = np.clip(scores, 0, 1000)
    return np.mean(scores)


def PE_sym(ref_preds, test_preds):
    scores_1 = test_preds / (1. - test_preds + 10 ** -6) - 1.
    scores_1 = np.clip(scores_1, 0, 1000)
    scores_2 = (1. - ref_preds) / (ref_preds + 10 ** -6) - 1.
    scores_2 = np.clip(scores_2, 0, 1000)
    return np.mean(scores_1) + np.mean(scores_2)


def Wasserstein(ref_preds, test_preds):
    return np.mean(test_preds) - np.mean(ref_preds)


def change_point_roc_auc(T_label, label, T_score, score, window_size):
    """
    T_label : numpy.array
        Ineces of time series observations and their labels.
    label   : numpy.array
        Labels of time series observations. 1 - change point, 0 - ordinary observation.
    T_score : numpy.array
        Indeces of change point detection score.
    score   : numpy.array
        Change point detection score. 0 - ordinary observation, high value - change point.
    window_size : int
        All observations with t < 2 * window_size after a change point are marked as 1
        and considered as a new collective change point.
        
    Returns
    -------
    auc : float
        ROC AUC score calculated based on change point detection score and new collective change points.
    """
    
    new_label = np.zeros(len(label))
    T_change = T_label[label == 1]
    for t in T_change:
        cond = (T_label-t < 2 * window_size) * (T_label-t >= 0)
        new_label[cond] = 1
    
    new_label = new_label[T_score]
    auc = roc_auc_score(new_label, score)
    
    return auc



import numpy as np
from matplotlib import pyplot as plt


def display(X, cps_true, score, cps_pred=None, plot_peak_height=10, s_max=10):

    T = np.arange(len(X))
    L = np.zeros(len(X))
    L[cps_true] = 1

    n = X.shape[1] + 1 if cps_pred is None else X.shape[1] + 2

    plt.figure(figsize=(12, n*2.5+0.25))

    for i in range(X.shape[1]):

        ax = X[:, i]
        d = 0.05 * (ax.max() - ax.min())

        plt.subplot(n, 1, i+1)
        plt.plot(T, ax, linewidth=2, label='Original signal', color='C0')
        for t in T[L == 1]:
            plt.plot([t]*2, [ax.min()-d, ax.max()+d], color='0', linestyle='--')
        plt.ylim(ax.min()-d, ax.max()+d)
        plt.xlim(0, T.max())
        plt.xticks(size=16)
        plt.yticks(size=16)
        plt.legend(loc='upper left', fontsize=16)
        plt.tight_layout()

    score_plot_ix = n if cps_pred is None else n - 1
    d = 0.05 * (score.max() - score.min())
    plt.subplot(n, 1, score_plot_ix)
    plt.plot(T, score, linewidth=3, label="Change point score", color='C3')
    for t in T[L == 1]:
        plt.plot([t]*2, [score.min()-d, score.max()+d], color='0', linestyle='--')
    plt.ylim(score.min()-d, score.max()+d)
    plt.xlim(0, T.max())
    plt.xticks(size=16)
    plt.yticks(size=16)
    plt.legend(loc='upper left', fontsize=16)
    plt.tight_layout()

    # display find peaks #todo refactoring
    if cps_pred is not None:
        plt.subplot(n, 1, n)
        L_pred = np.zeros(len(T))
        L_pred[cps_pred] = 1
        plt.plot(L_pred, linewidth=3, label="Change points", color='C4')
        for t in T[L == 1]:
            plt.plot([t]*2, [-0.05, 1.05], color='0', linestyle='--')
        plt.ylim(-0.05, 1.05)
        plt.xlim(0, T.max())
        plt.xticks(size=16)
        plt.yticks(size=16)
        plt.legend(loc='upper left', fontsize=16)

    plt.xlabel("Time", size=16)
    plt.tight_layout()



import torch.nn as nn


class MyNN(nn.Module):
    def __init__(self, n_inputs=1, n_hidden=10, dropout=0.):
        super(MyNN, self).__init__()
        self.fc1 = nn.Linear(n_inputs, n_hidden)
        self.act1 = nn.Sigmoid()
        self.fc2 = nn.Linear(n_hidden, 1)
        self.act2 = nn.Sigmoid()
    
    def forward(self, x):
        x = self.fc1(x)
        x = self.act1(x)
        x = self.fc2(x)
        x = self.act2(x)
        return x


class MyNNRegressor(nn.Module):
    def __init__(self, n_inputs=1, n_hidden=10, dropout=0):
        super(MyNNRegressor, self).__init__()
        self.fc1 = nn.Linear(n_inputs, n_hidden)
        self.act1 = nn.Sigmoid()
        self.fc2 = nn.Linear(n_hidden, 1)
    
    def forward(self, x):
        x = self.fc1(x)
        x = self.act1(x)
        x = self.fc2(x)
        return x


from .pr import find_peaks
from .pr import precision_recall_scores, precision_score, recall_score
from .pr import true_positive_rate, false_positive_rate
from .pr import precision_recall_curve, auc_score, pr_auc

__all__ = [
    'find_peaks', 
    'precision_recall_scores', 
    'precision_score', 
    'recall_score'
    'true_positive_rate', 
    'false_positive_rate'
    'precision_recall_curve', 
    'auc_score', 
    'pr_auc'
]


from abc import ABCMeta, abstractmethod
import numpy as np
import pandas as pd
from sklearn.metrics import roc_curve, roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from copy import deepcopy
from joblib import Parallel, delayed
from scipy import interpolate
from scipy.signal import argrelmax


# utils
def autoregression_matrix(X, periods=1, fill_value=0):
    shifted_x = [pd.DataFrame(X).shift(periods=i, fill_value=fill_value).values for i in range(periods)]
    X_auto = np.hstack(tuple(shifted_x))
    return X_auto

def reference_test(X, window_size=2, step=1):
    T = []
    reference = []
    test = []
    for i in range(2*window_size-1, len(X), step):
        T.append(i)
        reference.append(X[i-2*window_size+1:i-window_size+1])
        test.append(X[i-window_size+1:i+1])
    return np.array(T), np.array(reference), np.array(test)

def KL_score_unsym(ref_ratios, test_ratios):
    score = np.mean(np.log(test_ratios))
    return score

def KL_score(ref_ratios, test_ratios):
    score = KL_score_unsym(ref_ratios, test_ratios) + KL_score_unsym(1./test_ratios, 1./ref_ratios)
    return score

def PE_score_unsym(ref_ratios, test_ratios, alpha=0.):
    score = (-0.5 *       alpha  * np.mean(test_ratios**2)) + \
            (-0.5 * (1. - alpha) * np.mean(ref_ratios**2))  + np.mean(test_ratios) - 0.5
    return score

def PE_score(ref_ratios, test_ratios, alpha=0.):
    score = PE_score_unsym(ref_ratios, test_ratios, alpha)# - PE_score_unsym(test_ratios, ref_ratios, alpha)
    return score

def KL(ref_preds, test_preds):
    return np.mean(np.log(test_preds + 10**-3)) - np.mean(np.log(1. - test_preds + 10**-3))

def KL_sym(ref_preds, test_preds):
    return np.mean(np.log(test_preds + 10**-3))     - np.mean(np.log(1. - test_preds + 10**-3)) + \
           np.mean(np.log(1. - ref_preds + 10**-3)) - np.mean(np.log(ref_preds + 10**-3))

def JSD(ref_preds, test_preds):
    return np.log(2) + 0.5 * np.mean(np.log(test_preds + 10**-3)) + 0.5 * np.mean(np.log(1. - ref_preds + 10**-3))

def PE(ref_preds, test_preds):
    scores = test_preds / (1. - test_preds + 10**-6) - 1.
    scores = np.clip(scores, 0, 1000)
    return np.mean(scores)

def PE_sym(ref_preds, test_preds):
    scores_1 = test_preds / (1. - test_preds + 10**-6) - 1.
    scores_1 = np.clip(scores_1, 0, 1000)
    scores_2 = (1. - ref_preds) / (ref_preds + 10**-6) - 1.
    scores_2 = np.clip(scores_2, 0, 1000)
    return np.mean(scores_1) + np.mean(scores_2)



# Base
class ChangePointDetectionBase(metaclass=ABCMeta):
    
    def __init__(self, periods=1, window_size=100, step=1, n_runs=1):
        """
        Change point detection algorithm based on binary classififcation.

        Parameters:
        -----------
        periods: int
            Number of consecutive observations of a time series, considered as one input vector.
        window_size: int
            Number of consecutive observations of a time series in test and reference windows.
        step: int
            Algorithm estimates change point detection score for each <step> observation.
        n_runs: int
            Number of times, the binary classifier runs on each pair of test and reference windows.
        
        """

        self.periods = periods
        self.window_size = window_size
        self.step = step
        self.n_runs = n_runs
        
        
    def reference_test_predict(self, X_ref, X_test):
        """
        Estimate change point detection score for a pair of test and reference windows.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Retunrs:
        --------
        score: float
            Estimated change point detection score for a pair of window.
        """

        score = 0
        
        return score
    
    
    def reference_test_predict_n_times(self, X_ref, X_test):
        """
        Estimate change point detection score several times for  a pair of test and reference windows.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Returns:
        --------
        score: float
            Estimated average change point detection score for a pair of window.
        """

        scores = []
        for i in range(self.n_runs):
            ascore = self.reference_test_predict(X_ref, X_test)
            scores.append(ascore)
        
        return np.mean(scores)


    def posprocessing(self, T, T_score, score):
        """
        Interpolates and shifts a change point detection score, estimates peak positions.
        
        Parameters:
        -----------
        T: numpy.array
            A broader time-step interval
        T_score: numpy.array
            A time intervals of CPD scores
        score: numpy.array
            A CPD scores

        Returns:
        --------
        new_score: numpy.array
            Interpolated and shifted CPD scores.
        peaks: numpy.array
            Positions of peaks in the CPD scores.
        """

        inter = interpolate.interp1d(T_score-self.window_size, score, 
                                     kind='linear', fill_value=(0, 0), bounds_error=False)
        new_score = inter(T)
        peaks = argrelmax(new_score, order=self.window_size)[0]

        return new_score, peaks
        
    
    def predict(self, X):
        """
        Estimate change point detection score for a time series.

        Parameters:
        -----------
        X: numpy.ndarray
            Time series observation.

        Retunrs:
        --------
        T_score: numpy.array
            Array of timestamps.
        scores: numpy.array
            Estimated change point detection score.
        """
        
        X_auto = autoregression_matrix(X, periods=self.periods, fill_value=0)
        T, reference, test = reference_test(X_auto, window_size=self.window_size, step=1)
        
        scores = []
        T_score = []
        iters = range(0, len(reference), self.step)
        score = Parallel(n_jobs=-1)(delayed(self.reference_test_predict_n_times)(reference[i], test[i]) for i in iters)
        T_score = np.array([T[i] for i in iters])

        new_score, peaks = self.posprocessing(np.arange(len(X)), T_score, score)
        
        return new_score, peaks

    
    
    
# Classification
class ChangePointDetectionClassifier(ChangePointDetectionBase):
    
    def __init__(self, base_classifier=QuadraticDiscriminantAnalysis(), metric="KL_sym", 
                 periods=1, window_size=100, step=1, n_runs=1):
        """
        Change point detection algorithm based on binary classififcation.

        Parameters:
        -----------
        base_classifier: object
            Sklearn-like binary classifier.
        metric: string
            Name of the metric, that is used to measure the classifier quality and 
            considered as change point detection score. Default: "KL_sym".
        periods: int
            Number of consecutive observations of a time series, considered as one input vector.
        window_size: int
            Number of consecutive observations of a time series in test and reference windows.
        step: int
            Algorithm estimates change point detection score for each <step> observation.
        n_runs: int
            Number of times, the binary classifier runs on each pair of test and reference windows.
        
        """

        super().__init__(periods, window_size, step, n_runs)
        self.base_classifier = base_classifier
        self.metric = metric
        

    def densratio(self, y, alpha=10**-3):
        w = (y + alpha) / (1 - y + alpha)
        return w
        
        
    def reference_test_predict(self, X_ref, X_test):
        """
        Estimate change point detection score for a pair of test and reference windows.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Retunrs:
        --------
        score: float
            Estimated change point detection score for a pair of window.
        """

        y_ref = np.zeros(len(X_ref))
        y_test = np.ones(len(X_test))
        X = np.vstack((X_ref, X_test))
        y = np.hstack((y_ref, y_test))

        X_train, X_test, y_train, y_test = train_test_split(X, y, 
                                                            test_size=0.5, 
                                                            stratify=y, 
                                                            random_state=np.random.randint(0, 1000))
        
        classifier = deepcopy(self.base_classifier)
        classifier.fit(X_train, y_train)
        y_pred = classifier.predict_proba(X_test)[:, 1]
        
        ref_preds = y_pred[y_test == 0]
        test_preds = y_pred[y_test == 1]
        ratios = self.densratio(y_pred)
        ref_ratios = ratios[y_test == 0]
        test_ratios = ratios[y_test == 1]
        
        if self.metric == "KL":
            score = KL(ref_preds, test_preds)
        elif self.metric == "KL_sym":
            score = KL_sym(ref_preds, test_preds)
        elif self.metric == "JSD":
            score = JSD(ref_preds, test_preds)
        elif self.metric == "PE":
            score = PE(ref_preds, test_preds)
        elif self.metric == "PE_sym":
            score = PE_sym(ref_preds, test_preds)
        elif self.metric == "ROCAUC":
            score = 2 * (roc_auc_score(y_test, y_pred) - 0.5)
        else:
            score = 0
        
        return score
    


       
# NN RuLSIF
class ChangePointDetectionRuLSIF(ChangePointDetectionBase):
    
    def __init__(self, base_regressor, metric="PE", periods=1, window_size=100, step=1, n_runs=1):
        """
        Change point detection algorithm based on RuLSIF regressor.

        Parameters:
        -----------
        base_regressor: object
            Sklearn-like regressor with RuLSIF loss function.
        metric: string
            Name of the metric, that is used to measure the classifier quality and 
            considered as change point detection score. Default: "PE".
        periods: int
            Number of consecutive observations of a time series, considered as one input vector.
        window_size: int
            Number of consecutive observations of a time series in test and reference windows.
        step: int
            Algorithm estimates change point detection score for each <step> observation.
        n_runs: int
            Number of times, the binary classifier runs on each pair of test and reference windows.
        
        """

        super().__init__(periods, window_size, step, n_runs)
        self.base_regressor = base_regressor
        self.metric = metric


    def one_side_predict(self, X_train, X_test, y_train, y_test):
        """
        Fit a regressor on a pair of the train sample and make a prediction on the test.

        Parameters:
        -----------
        X_train: numpy.ndarray
            Matrix of train observations.
        X_test: numpy.ndarray
            Matrix of test observations.
        y_train: numpy.array
            Array of train labels.
        y_test: numpy.array
            Array of test labels.

        Retunrs:
        --------
        score: float
            Estimated change point detection score for the test.
        """

        reg = deepcopy(self.base_regressor)
        reg.fit(X_train, y_train)
        ratios = reg.predict(X_test)
        ref_ratios = ratios[y_test == 0]
        test_ratios = ratios[y_test == 1]
        if self.metric == "PE":
            score = 0.5 * np.mean(test_ratios) - 0.5 # PE_score_unsym(ref_ratios, test_ratios, classifier_1.alpha) - PE_score_unsym(test_ratios, ref_ratios, classifier_1.alpha)
        else:
            score = 0

        return score
        
        
    def reference_test_predict(self, X_ref, X_test):
        """
        Estimate change point detection score for a pair of test and reference windows.

        Parameters:
        -----------
        X_ref: numpy.ndarray
            Matrix of reference observations.
        X_test: numpy.ndarray
            Matrix of test observations.

        Retunrs:
        --------
        score: float
            Estimated change point detection score for a pair of window.
        """

        y_ref = np.zeros(len(X_ref))
        y_test = np.ones(len(X_test))
        X = np.vstack((X_ref, X_test))
        y = np.hstack((y_ref, y_test))

        X_train, X_test, y_train, y_test = train_test_split(X, y, 
                                                            test_size=0.5, 
                                                            stratify=y, 
                                                            random_state=np.random.randint(0, 1000))
        
        score_right = self.one_side_predict(X_train, X_test, y_train, y_test)
        score_left = self.one_side_predict(X_train, X_test, 1-y_train, 1-y_test)
        score = score_right + score_left
        
        return score




import numpy as np
from matplotlib import pyplot as plt


def dataset1(period=200, N_tot=1000):
    mu = 0
    sigma = 1.
    N = 1
    
    T = [0, 1]
    X = [np.random.normal(mu, sigma, 1)[0], np.random.normal(mu, sigma, 1)[0]]
    
    for i in range(2, N_tot):
        if i % period == 0:
            N += 1
            mu += 0.5 * N
        T += [i]
        ax = 0.6 * X[i - 1] - 0.5 * X[i - 2] + np.random.normal(mu, sigma, 1)[0]
        X += [ax]
    return np.array(X).reshape(-1, 1)


def plot(X, Y, title, x_label, y_label):
    plt.figure(figsize=(12, 3.))
    plt.plot(X, Y, linewidth=3)
    
    plt.xlabel(x_label, size=14)
    plt.ylabel(y_label, size=14)
    plt.grid(b=1)
    plt.title(title, size=14)
    # plt.legend(loc='best')
    plt.tight_layout()
    plt.xlim(0)
    # plt.show()


def plot2(title, x_label, y_label):
    fig = plt.figure(figsize=(15, 6))
    ax = fig.add_subplot(1, 1, 1)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.grid(b=1)
    # ax.plot([1, 2, 3], [2, 4, 6])
    ax.set_title(title)
    plt.tight_layout()
    # ax.set_xlim(0)
    return ax


def draw_projection(ax, score, period, ws, n):
    min_height, max_height = np.min(score), np.max(score)
    for line in range(1, n):
        current_period = period * line
        c_ws = current_period + ws
        ax.plot([current_period, current_period], [min_height, max_height], '--', alpha=1, lw=1.5, c='k')
        ax.plot([c_ws, c_ws], [min_height, max_height], '-.', alpha=0.5, lw=1, c='k')
    return ax


def SMA(scores, N):
    new_scores = []
    for i in range(0, len(scores)):
        s = i - N if i - N >= 0 else 0
        new_scores.append(np.mean(scores[s:i + 1], axis=0))
    return np.array(new_scores)


def EMA(scores, N, smooth):
    alpha = smooth / (1. + N)
    new_scores = [scores[0]]
    for i in range(1, len(scores)):
        new_score = alpha * scores[i] + (1 - alpha) * new_scores[i - 1]
        new_scores.append(new_score)
    return new_scores


def cum_sum(scores):
    new_scores = [0.]
    for i in range(1, len(scores)):
        new_scores.append(np.maximum(0., new_scores[i - 1] + scores[i]))
    return new_scores


def generate_simplest_data(period=200, N=2):
    X = []
    for i in range(N):
        temp = [i for j in range(period)]
        X += temp
    return np.array(X).reshape(-1, 1)


def generate_normal_simple_data(period=200, N_tot=1000):
    mu = 0
    sigma = 0.1
    N = 1
    
    T = [0, 1]
    X = [np.random.normal(mu, sigma, 1)[0], np.random.normal(mu, sigma, 1)[0]]
    
    for i in range(2, N_tot):
        if i % period == 0:
            N += 1
            mu += 0.5 * N
        T += [i]
        ax = 0.6 * X[i - 1] - 0.5 * X[i - 2] + np.random.normal(mu, sigma, 1)[0]
        X += [ax]
    return np.array(X).reshape(-1, 1)



from .algorithms import OnlineNNClassifier, OnlineNNRuLSIF
from .cpdc import ChangePointDetectionClassifier, ChangePointDetectionRuLSIF
from .models import GBDTRuLSIFRegressor
from .models import RegressionNetwork, NNRuLSIFRegressor
from .models import ClassificationNetwork, NNClassifier


__all__ = [
    'OnlineNNClassifier', 
    'OnlineNNRuLSIF', 
    'ChangePointDetectionClassifier', 
    'ChangePointDetectionRuLSIF'
    'ClassificationNetwork', 
    'NNClassifier'
    'GBDTRuLSIFRegressor', 
    'RegressionNetwork', 
    'NNRuLSIFRegressor'
]



import numpy as np
import pandas as pd

from scipy import interpolate


def unified_score(T: np.ndarray, T_score: np.ndarray, score: np.ndarray) -> np.ndarray:
    """
    Interpolates a CPD score of T_score interval onto T interval
    Parameters
    ----------
    T: A broader time-step interval
    T_score: A time intervals of CPD scores
    score: A CPD scores

    Returns
    -------
    Interpolated CPD scores
    """
    inter = interpolate.interp1d(T_score, score, kind='previous', fill_value=(0, 0), bounds_error=False)
    uni_score = inter(T)
    return uni_score


def autoregression_matrix(X, periods=1, fill_value=0):
    shifted_x = [pd.DataFrame(X).shift(periods=i, fill_value=fill_value).values for i in range(periods)]
    X_auto = np.hstack(tuple(shifted_x))
    return X_auto


def reference_test(X, window_size=2, step=1):
    T = []
    reference = []
    test = []
    for i in range(2 * window_size - 1, len(X), step):
        T.append(i)
        reference.append(X[i - 2 * window_size + 1:i - window_size + 1])
        test.append(X[i - window_size + 1:i + 1])
    return np.array(T), np.array(reference), np.array(test)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="roerich",  # Replace with your own username
    version="0.3.0",
    author="Mikhail Hushchyn, Kenenbek Arzymatov",
    author_email="hushchyn.mikhail@gmail.com, kenenbek@gmail.com",
    description="Roerich is a python library for online and offline change point detection in time series data based on machine learning.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HSE-LAMBDA/roerich",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'numpy>=1.23.0',
        'pandas>=1.4.2',
        'scipy>=1.8.1',
        'scikit-learn>=1.1.1',
        'joblib>=1.1.0',
        'matplotlib>=3.5.2',
        'torch>=1.11.0',
        'torchaudio>=0.11.0',
        'torchvision>=0.12.0'
    ],
)



import numpy as np
import pandas as pd
from scipy.signal import argrelmax

def find_peaks(score, order):
    """
    Searches peaks in CPD score and estimates their heights.

    Parameters:
    -----------
    score: array-like
        Change point detection score.
    order: ind
        The number of points to the left and right of the peak to compare.

    Returns:
    --------
    cps_pred: numpy.array
        Indices of the detected change points.
    cps_score_pred: numpy.array
        Change point detection score for the detected indices of the change points.
    """
    
    cps_pred = argrelmax(score, order=order)[0]
    cps_score_pred = score[cps_pred]
    
    return cps_pred, cps_score_pred


def precision_recall_scores(cps_true, cps_pred, window=20):
    """
    Calculates precision and recall scores.

    Parameters:
    -----------
    cps_true: array-like
        True indices of change points. Example: [0, 100, 200, 300].
    cps_pred: array-like
        Detected indices of change points. Example: [0, 102, 195, 298, 303].
    window: int
        Maximum allowed distance between true and predicted change points.

    Retunrs:
    --------
    precision: float
        Precision score value.
    recall: float
        Recall score value.
    """

    n_cr = 0
    n_cp = len(cps_true)
    n_al = len(cps_pred)
    
    if n_al == 0: 
        return 0, 0
    
    if n_cp == 0:
        return 0, 1

    is_used = []
    for atrue in cps_true:
        for apred in cps_pred:
            if (np.abs(apred - atrue) <= window) and (apred not in is_used):
                n_cr += 1
                is_used.append(apred)
                break

    tpr = n_cr / n_cp
    fpr = (n_al - n_cr) / n_al

    recall = tpr
    precision = 1 - fpr
    
    return precision, recall



def precision_score(cps_true, cps_pred, window=20):
    """
    Calculates precision score.

    Parameters:
    -----------
    cps_true: array-like
        True indices of change points. Example: [0, 100, 200, 300].
    cps_pred: array-like
        Detected indices of change points. Example: [0, 102, 195, 298, 303].
    window: int
        Maximum allowed distance between true and predicted change points.

    Retunrs:
    --------
    precision: float
        Precision score value.
    """

    precision, recall = precision_recall_scores(cps_true, cps_pred, window=20)
    return precision



def recall_score(cps_true, cps_pred, window=20):
    """
    Calculates recall score.

    Parameters:
    -----------
    cps_true: array-like
        True indices of change points. Example: [0, 100, 200, 300].
    cps_pred: array-like
        Detected indices of change points. Example: [0, 102, 195, 298, 303].
    window: int
        Maximum allowed distance between true and predicted change points.

    Retunrs:
    --------
    recall: float
        Recall score value.
    """

    precision, recall = precision_recall_scores(cps_true, cps_pred, window=20)
    return recall



def true_positive_rate(cps_true, cps_pred, window=20):
    """
    Calculates true positive rate.

    Parameters:
    -----------
    cps_true: array-like
        True indices of change points. Example: [0, 100, 200, 300].
    cps_pred: array-like
        Detected indices of change points. Example: [0, 102, 195, 298, 303].
    window: int
        Maximum allowed distance between true and predicted change points.

    Retunrs:
    --------
    tpr: float
        True positive rate value.
    """

    precision, recall = precision_recall_scores(cps_true, cps_pred, window=20)
    return recall



def false_positive_rate(cps_true, cps_pred, window=20):
    """
    Calculates false positive rate.

    Parameters:
    -----------
    cps_true: array-like
        True indices of change points. Example: [0, 100, 200, 300].
    cps_pred: array-like
        Detected indices of change points. Example: [0, 102, 195, 298, 303].
    window: int
        Maximum allowed distance between true and predicted change points.

    Retunrs:
    --------
    fpr: float
        False positive rate value.
    """

    precision, recall = precision_recall_scores(cps_true, cps_pred, window=20)
    return 1 - precision



def precision_recall_curve(cps_true, cps_pred, cps_score_pred=None, window=20):
    """
    Calculates precision-recall curve.

    Parameters:
    -----------
    cps_true: array-like
        True indices of change points. Example: [0, 100, 200, 300].
    cps_pred: array-like
        Detected indices of change points. Example: [0, 102, 195, 298, 303].
    cps_score_pred: array-like
        Change point detection score for the detected indices of the change points. 
        If 'None', then the score for all change points is taken equal to 1.
    window: int
        Maximum allowed distance between true and predicted change points.

    Retunrs:
    --------
    thresholds: array-like
        Different thresholds for CPD scores of the detected change points.
    precision: array-like
        Precision score values according to the thresholds.
    recall: array-like
        Recall score values according to the thresholds.
    """

    if cps_score_pred is None:
        cps_score_pred = np.ones(cps_pred.shape)

    thresholds = np.unique(cps_score_pred)
    
    data = []
    for thr in thresholds:
        
        cps_thr = cps_pred[cps_score_pred > thr]
        precision, recall = precision_recall_scores(cps_true, cps_pred, window)
        data.append([thr, precision, recall])
        
    data.insert(0, [-999, 0.0, 1.0])
    data.append([999, 1.0, 0.0])
    data = np.array(data)

    return data[:, 0], data[:, 1], data[:, 2]



def auc_score(thresholds, precision, recall):
    """
    Calculates auc.

    Parameters:
    -----------
    thresholds: array-like
        Different thresholds for CPD scores of the detected change points.
    precision: array-like
        Precision score values according to the thresholds.
    recall: array-like
        Recall score values according to the thresholds.

    Returns:
    --------
    auc: float
        The precision-recall curve auc value.
    """
    
    sorted_inds = thresholds.argsort()
    auc = np.abs(np.trapz(precision[sorted_inds], recall[sorted_inds]))
    
    return auc



def pr_auc(cps_true, cps_pred, cps_score_pred=None, window=20):
    """
    Calculates precision-recall curve auc.

    Parameters:
    -----------
    cps_true: array-like
        True indices of change points. Example: [0, 100, 200, 300].
    cps_pred: array-like
        Detected indices of change points. Example: [0, 102, 195, 298, 303].
    cps_score_pred: array-like
        Change point detection score for the detected indices of the change points. 
        If 'None', then the score for all change points is taken equal to 1.
    window: int
        Maximum allowed distance between true and predicted change points.

    Retunrs:
    --------
    auc: float
        The precision-recall curve auc value.
    """

    thresholds, precision, recall = precision_recall_curve(cps_true, cps_pred, cps_score_pred, window)
    auc = auc_score(thresholds, precision, recall)

    return auc




import numpy as np
import collections
import itertools
from abc import ABCMeta, abstractmethod


class BaseScaler(metaclass=ABCMeta):
    def __init__(self, N):
        self.N = N
        self.mu = None
        self.sigma = None
    
    @abstractmethod
    def fit(self, X):
        pass
    
    def transform(self, X):
        return (X - self.mu) / self.sigma
    
    def fit_transform(self, X):
        self.fit(X)
        return self.transform(X)


class SmaScaler(BaseScaler):
    def fit(self, X):
        if self.sigma is None:
            self.mu = np.mean(X, axis=0)
            self.sigma = np.std(X, axis=0)
        else:
            x = np.mean(X, axis=0)
            self.sigma = np.sqrt(self.sigma ** 2 * (self.N - 2) / (self.N - 1) + (x - self.mu) ** 2 / self.N)
            self.mu += (x - self.mu) / self.N


class EmaScaler(BaseScaler):
    def __init__(self, N, smooth):
        super().__init__(N)
        self.alpha = smooth / (1. + self.N)
    
    def fit(self, X):
        if self.sigma is None:
            self.mu = np.mean(X, axis=0)
            self.sigma = np.std(X, axis=0)
        else:
            x = np.mean(X, axis=0)
            self.sigma = (1 - self.alpha) * (self.sigma + self.alpha * (x - self.mu) ** 2)
            self.mu = self.alpha * x + (1 - self.alpha) * self.mu


class SmaScalerCache(BaseScaler):
    def __init__(self, N):
        super(SmaScalerCache, self).__init__(N)
        self.cache = []
    
    def fit(self, X):
        if len(self.cache) + len(X) > self.N:
            over = len(self.cache) + len(X) - self.N
            over = min(over, len(self.cache))
            self.cache = self.cache[-over:]
        self.cache.extend(X)
        
        self.mu = np.mean(self.cache, axis=0)
        self.sigma = np.std(self.cache, axis=0)


class EmaScalerCache(EmaScaler):  # todo
    def __init__(self, N, smooth):
        super().__init__(N, smooth)
        self.cache = collections.deque()
    
    def fit(self, X):
        if self.sigma is None:
            self.mu = np.mean(X, axis=0)
            self.sigma = np.std(X, axis=0)
        else:
            x = np.mean(X, axis=0)
            self.sigma = (1 - self.alpha) * (self.sigma + self.alpha * (x - self.mu) ** 2)
            self.mu = self.alpha * x + (1 - self.alpha) * self.mu


class MockScaler(BaseScaler):
    def fit(self, X):
        pass
    
    def transform(self, X):
        return X
